@@ -1,6 +1,6 @@
 ;;; tramp-adb.el --- Functions for calling Android Debug Bridge from Tramp  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2011-2018 Free Software Foundation, Inc.
+;; Copyright (C) 2011-2019 Free Software Foundation, Inc.
 
 ;; Author: Jürgen Hötzel <juergen@archlinux.org>
 ;; Keywords: comm, processes
@@ -53,7 +53,7 @@ It is used for TCP/IP devices."
   "When this method name is used, forward all calls to Android Debug Bridge.")
 
 (defcustom tramp-adb-prompt
-  "^\\(?:[[:digit:]]*|?\\)?\\(?:[[:alnum:]\e;[]*@?[[:alnum:]]*[^#\\$]*\\)?[#\\$][[:space:]]"
+  "^[[:digit:]]*|?[[:alnum:]\e;[]*@?[[:alnum:]]*[^#\\$]*[#\\$][[:space:]]"
   "Regexp used as prompt in almquist shell."
   :type 'string
   :version "24.4"
@@ -78,8 +78,8 @@ It is used for TCP/IP devices."
 (tramp--with-startup
  (add-to-list 'tramp-methods
 	      `(,tramp-adb-method
-	        (tramp-tmpdir "/data/local/tmp")
-                (tramp-default-port 5555)))
+	        (tramp-tmpdir            "/data/local/tmp")
+                (tramp-default-port      5555)))
 
  (add-to-list 'tramp-default-host-alist `(,tramp-adb-method nil ""))
 
@@ -88,7 +88,7 @@ It is used for TCP/IP devices."
 
 ;;;###tramp-autoload
 (defconst tramp-adb-file-name-handler-alist
-  '((access-file . ignore)
+  '((access-file . tramp-handle-access-file)
     (add-name-to-file . tramp-handle-add-name-to-file)
     ;; `byte-compiler-base-file-name' performed by default handler.
     ;; `copy-directory' performed by default handler.
@@ -153,7 +153,7 @@ It is used for TCP/IP devices."
     (set-file-selinux-context . ignore)
     (set-file-times . tramp-adb-handle-set-file-times)
     (set-visited-file-modtime . tramp-handle-set-visited-file-modtime)
-    (shell-command . tramp-adb-handle-shell-command)
+    (shell-command . tramp-handle-shell-command)
     (start-file-process . tramp-handle-start-file-process)
     (substitute-in-file-name . tramp-handle-substitute-in-file-name)
     (temporary-file-directory . tramp-handle-temporary-file-directory)
@@ -191,38 +191,14 @@ pass to the OPERATION."
 ;;;###tramp-autoload
 (defun tramp-adb-parse-device-names (_ignore)
   "Return a list of (nil host) tuples allowed to access."
-  (with-timeout (10)
-    (with-temp-buffer
-      ;; `call-process' does not react on timer under MS Windows.
-      ;; That's why we use `start-process'.
-      ;; We don't know yet whether we need a user or host name for the
-      ;; connection vector.  We assume we don't, it will be OK in most
-      ;; of the cases.  Otherwise, there might be an additional trace
-      ;; buffer, which doesn't hurt.
-      (let ((p (start-process
-		tramp-adb-program (current-buffer) tramp-adb-program "devices"))
-	    (v (make-tramp-file-name :method tramp-adb-method))
-	    result)
-	(tramp-message v 6 "%s" (mapconcat 'identity (process-command p) " "))
-	(process-put p 'adjust-window-size-function 'ignore)
-	(set-process-query-on-exit-flag p nil)
-	(while (process-live-p p)
-	  (accept-process-output p 0.1))
-	(accept-process-output p 0.1)
-	(tramp-message v 6 "\n%s" (buffer-string))
-	(goto-char (point-min))
-	(while (search-forward-regexp "^\\(\\S-+\\)[[:space:]]+device$" nil t)
-	  (push (list nil (match-string 1)) result))
-
-	;; Replace ":" by "#".
-	(mapc
-	 (lambda (elt)
-	   (setcar
-	    (cdr elt)
-	    (replace-regexp-in-string
-	     ":" tramp-prefix-port-format (car (cdr elt)))))
-	 result)
-	result))))
+  (delq nil
+	(mapcar
+	 (lambda (line)
+	   (when (string-match "^\\(\\S-+\\)[[:space:]]+device$" line)
+	     ;; Replace ":" by "#".
+	     `(nil ,(replace-regexp-in-string
+		     ":" tramp-prefix-port-format (match-string 1 line)))))
+	 (tramp-process-lines nil tramp-adb-program "devices"))))
 
 (defun tramp-adb-handle-file-system-info (filename)
   "Like `file-system-info' for Tramp files."
@@ -256,7 +232,7 @@ pass to the OPERATION."
    ;; Preserve trailing "/".
   (funcall
    (if (string-equal (file-name-nondirectory filename) "")
-       'file-name-as-directory 'identity)
+       #'file-name-as-directory #'identity)
    (with-parsed-tramp-file-name (expand-file-name filename) nil
      (tramp-make-tramp-file-name
       v
@@ -281,17 +257,15 @@ pass to the OPERATION."
 	      (setq thisstep (pop steps))
 	      (tramp-message
 	       v 5 "Check %s"
-	       (mapconcat 'identity
-			  (append '("") (reverse result) (list thisstep))
-			  "/"))
+	       (string-join
+		(append '("") (reverse result) (list thisstep)) "/"))
 	      (setq symlink-target
 		    (tramp-compat-file-attribute-type
 		     (file-attributes
 		      (tramp-make-tramp-file-name
-		       v (mapconcat 'identity
-				    (append
-				     '("") (reverse result) (list thisstep))
-				    "/")))))
+		       v
+		       (string-join
+			(append '("") (reverse result) (list thisstep)) "/")))))
 	      (cond ((string= "." thisstep)
 		     (tramp-message v 5 "Ignoring step `.'"))
 		    ((string= ".." thisstep)
@@ -326,9 +300,9 @@ pass to the OPERATION."
 	    ;; Combine list to form string.
 	    (setq result
 		  (if result
-		      (mapconcat 'identity (cons "" result) "/")
+		      (string-join (cons "" result) "/")
 		    "/"))
-	    (when (and is-dir (or (string= "" result)
+	    (when (and is-dir (or (string-empty-p result)
 				  (not (string= (substring result -1) "/"))))
 	      (setq result (concat result "/"))))
 
@@ -468,10 +442,10 @@ pass to the OPERATION."
   "Almquist shell can't handle multiple arguments.
 Convert (\"-al\") to (\"-a\" \"-l\").  Remove arguments like \"--dired\"."
   (split-string
-   (apply 'concat
+   (apply #'concat
 	  (mapcar (lambda (s)
 		    (replace-regexp-in-string
-		     "\\(.\\)"  " -\\1" (replace-regexp-in-string "^-" "" s)))
+		     "\\(.\\)" " -\\1" (replace-regexp-in-string "^-" "" s)))
 		  ;; FIXME: Warning about removed switches (long and non-dash).
 		  (delq nil
 			(mapcar
@@ -500,10 +474,10 @@ Emacs dired can't find files."
 	    (sort
 	     lines
 	     (if sort-by-time
-		 'tramp-adb-ls-output-time-less-p
-	       'tramp-adb-ls-output-name-less-p))))
+		 #'tramp-adb-ls-output-time-less-p
+	       #'tramp-adb-ls-output-name-less-p))))
       (delete-region (point-min) (point-max))
-      (insert "  " (mapconcat 'identity sorted-lines "\n  ")))
+      (insert "  " (string-join sorted-lines "\n  ")))
     ;; Add final newline.
     (goto-char (point-max))
     (unless (bolp) (insert "\n"))))
@@ -512,9 +486,9 @@ Emacs dired can't find files."
   "Sort \"ls\" output by time, descending."
   (let (time-a time-b)
     (string-match tramp-adb-ls-date-regexp a)
-    (setq time-a (apply 'encode-time (parse-time-string (match-string 0 a))))
+    (setq time-a (apply #'encode-time (parse-time-string (match-string 0 a))))
     (string-match tramp-adb-ls-date-regexp b)
-    (setq time-b (apply 'encode-time (parse-time-string (match-string 0 b))))
+    (setq time-b (apply #'encode-time (parse-time-string (match-string 0 b))))
     (time-less-p time-b time-a)))
 
 (defun tramp-adb-ls-output-name-less-p (a b)
@@ -590,7 +564,7 @@ Emacs dired can't find files."
 	    (delq
 	     nil
 	     (mapcar
-	      (lambda (l) (and (not (string-match-p  "^[[:space:]]*$" l)) l))
+	      (lambda (l) (and (not (string-match-p "^[[:space:]]*$" l)) l))
 	      (split-string (buffer-string) "\n")))))))))))
 
 (defun tramp-adb-handle-file-local-copy (filename)
@@ -610,9 +584,7 @@ Emacs dired can't find files."
 	  (ignore-errors (delete-file tmpfile))
 	  (tramp-error
 	   v 'file-error "Cannot make local copy of file `%s'" filename))
-	(set-file-modes
-	 tmpfile
-	 (logior (or (file-modes filename) 0) (string-to-number "0400" 8))))
+	(set-file-modes tmpfile (logior (or (file-modes filename) 0) #o0400)))
       tmpfile)))
 
 (defun tramp-adb-handle-file-writable-p (filename)
@@ -658,11 +630,9 @@ But handle the case, if the \"test\" command is not available."
 	   (tmpfile (tramp-compat-make-temp-file filename)))
       (when (and append (file-exists-p filename))
 	(copy-file filename tmpfile 'ok)
-	(set-file-modes
-	 tmpfile
-	 (logior (or (file-modes tmpfile) 0) (string-to-number "0600" 8))))
+	(set-file-modes tmpfile (logior (or (file-modes tmpfile) 0) #o0600)))
       (tramp-run-real-handler
-       'write-region (list start end tmpfile append 'no-message lockname))
+       #'write-region (list start end tmpfile append 'no-message lockname))
       (with-tramp-progress-reporter
         v 3 (format-message
              "Moving tmp file `%s' to `%s'" tmpfile filename)
@@ -705,12 +675,22 @@ But handle the case, if the \"test\" command is not available."
 			(tramp-compat-time-equal-p time tramp-time-doesnt-exist)
 			(tramp-compat-time-equal-p time tramp-time-dont-know))
 		    (current-time)
-		  time)))
+		  time))
+	  (quoted-name (tramp-shell-quote-argument localname)))
+      ;; Older versions of toybox 'touch' mishandle nanoseconds and/or
+      ;; trailing "Z", so fall back on plain seconds if nanoseconds+Z
+      ;; fails.  Also, fall back on old POSIX 'touch -t' if 'touch -d'
+      ;; (introduced in POSIX.1-2008) fails.
       (tramp-adb-send-command-and-check
-       ;; Use shell arithmetic because of Emacs integer size limit.
-       v (format "touch -t $(( %d * 65536 + %d )) %s"
-		 (car time) (cadr time)
-		 (tramp-shell-quote-argument localname))))))
+       v (format (concat "touch -d %s %s 2>/dev/null || "
+			 "touch -d %s %s 2>/dev/null || "
+			 "touch -t %s %s")
+		 (format-time-string "%Y-%m-%dT%H:%M:%S.%NZ" time t)
+		 quoted-name
+		 (format-time-string "%Y-%m-%dT%H:%M:%S" time t)
+		 quoted-name
+		 (format-time-string "%Y%m%d%H%M.%S" time t)
+		 quoted-name)))))
 
 (defun tramp-adb-handle-copy-file
   (filename newname &optional ok-if-already-exists keep-date
@@ -845,7 +825,7 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
   (with-parsed-tramp-file-name default-directory nil
     (let (command input tmpinput stderr tmpstderr outbuf ret)
       ;; Compute command.
-      (setq command (mapconcat 'tramp-shell-quote-argument
+      (setq command (mapconcat #'tramp-shell-quote-argument
 			       (cons program args) " "))
       ;; Determine input.
       (if (null infile)
@@ -940,80 +920,6 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
 	  (keyboard-quit)
 	ret))))
 
-(defun tramp-adb-handle-shell-command
-  (command &optional output-buffer error-buffer)
-  "Like `shell-command' for Tramp files."
-  (let* ((asynchronous (string-match-p "[ \t]*&[ \t]*\\'" command))
-	 ;; We cannot use `shell-file-name' and `shell-command-switch',
-	 ;; they are variables of the local host.
-	 (args (list "sh" "-c" (substring command 0 asynchronous)))
-	 current-buffer-p
-	 (output-buffer
-	  (cond
-	   ((bufferp output-buffer) output-buffer)
-	   ((stringp output-buffer) (get-buffer-create output-buffer))
-	   (output-buffer
-	    (setq current-buffer-p t)
-	    (current-buffer))
-	   (t (get-buffer-create
-	       (if asynchronous
-		   "*Async Shell Command*"
-		 "*Shell Command Output*")))))
-	 (error-buffer
-	  (cond
-	   ((bufferp error-buffer) error-buffer)
-	   ((stringp error-buffer) (get-buffer-create error-buffer))))
-	 (buffer
-	  (if (and (not asynchronous) error-buffer)
-	      (with-parsed-tramp-file-name default-directory nil
-		(list output-buffer (tramp-make-tramp-temp-file v)))
-	    output-buffer))
-	 (p (get-buffer-process output-buffer)))
-
-    ;; Check whether there is another process running.  Tramp does not
-    ;; support 2 (asynchronous) processes in parallel.
-    (when p
-      (if (yes-or-no-p "A command is running.  Kill it? ")
-	  (ignore-errors (kill-process p))
-	(tramp-user-error p "Shell command in progress")))
-
-    (if current-buffer-p
-	(progn
-	  (barf-if-buffer-read-only)
-	  (push-mark nil t))
-      (with-current-buffer output-buffer
-	(setq buffer-read-only nil)
-	(erase-buffer)))
-
-    (if (and (not current-buffer-p) (integerp asynchronous))
-	(prog1
-	    ;; Run the process.
-	    (apply 'start-file-process "*Async Shell*" buffer args)
-	  ;; Display output.
-	  (pop-to-buffer output-buffer)
-	  (setq mode-line-process '(":%s"))
-	  (shell-mode))
-
-      (prog1
-	  ;; Run the process.
-	  (apply 'process-file (car args) nil buffer nil (cdr args))
-	;; Insert error messages if they were separated.
-	(when (listp buffer)
-	  (with-current-buffer error-buffer
-	    (insert-file-contents (cadr buffer)))
-	  (delete-file (cadr buffer)))
-	(if current-buffer-p
-	    ;; This is like exchange-point-and-mark, but doesn't
-	    ;; activate the mark.  It is cleaner to avoid activation,
-	    ;; even though the command loop would deactivate the mark
-	    ;; because we inserted text.
-	    (goto-char (prog1 (mark t)
-			 (set-marker (mark-marker) (point)
-				     (current-buffer))))
-	  ;; There's some output, display it.
-	  (when (with-current-buffer output-buffer (> (point-max) (point-min)))
-	    (display-message-or-buffer output-buffer)))))))
-
 ;; We use BUFFER also as connection buffer during setup.  Because of
 ;; this, its original contents must be saved, and restored once
 ;; connection has been setup.
@@ -1026,31 +932,30 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
 	    (command (plist-get args :command))
 	    (coding (plist-get args :coding))
 	    (noquery (plist-get args :noquery))
-	    (stop (plist-get args :stop))
 	    (connection-type (plist-get args :connection-type))
 	    (filter (plist-get args :filter))
 	    (sentinel (plist-get args :sentinel))
 	    (stderr (plist-get args :stderr)))
 	(unless (stringp name)
-	  (signal 'wrong-type-argument (list 'stringp name)))
+	  (signal 'wrong-type-argument (list #'stringp name)))
 	(unless (or (null buffer) (bufferp buffer) (stringp buffer))
-	  (signal 'wrong-type-argument (list 'stringp buffer)))
+	  (signal 'wrong-type-argument (list #'stringp buffer)))
 	(unless (consp command)
-	  (signal 'wrong-type-argument (list 'consp command)))
+	  (signal 'wrong-type-argument (list #'consp command)))
 	(unless (or (null coding)
 		    (and (symbolp coding) (memq coding coding-system-list))
 		    (and (consp coding)
 			 (memq (car coding) coding-system-list)
 			 (memq (cdr coding) coding-system-list)))
-	  (signal 'wrong-type-argument (list 'symbolp coding)))
+	  (signal 'wrong-type-argument (list #'symbolp coding)))
 	(unless (or (null connection-type) (memq connection-type '(pipe pty)))
-	  (signal 'wrong-type-argument (list 'symbolp connection-type)))
+	  (signal 'wrong-type-argument (list #'symbolp connection-type)))
 	(unless (or (null filter) (functionp filter))
-	  (signal 'wrong-type-argument (list 'functionp filter)))
+	  (signal 'wrong-type-argument (list #'functionp filter)))
 	(unless (or (null sentinel) (functionp sentinel))
-	  (signal 'wrong-type-argument (list 'functionp sentinel)))
+	  (signal 'wrong-type-argument (list #'functionp sentinel)))
 	(unless (or (null stderr) (bufferp stderr) (stringp stderr))
-	  (signal 'wrong-type-argument (list 'stringp stderr)))
+	  (signal 'wrong-type-argument (list #'stringp stderr)))
 
 	(let* ((buffer
 		(if buffer
@@ -1060,17 +965,15 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
 	       (program (car command))
 	       (args (cdr command))
 	       (command
-		(format "cd %s; %s"
+		(format "cd %s && exec %s"
 			(tramp-shell-quote-argument localname)
-			(mapconcat 'tramp-shell-quote-argument
+			(mapconcat #'tramp-shell-quote-argument
 				   (cons program args) " ")))
 	       (tramp-process-connection-type
 		(or (null program) tramp-process-connection-type))
 	       (bmp (and (buffer-live-p buffer) (buffer-modified-p buffer)))
 	       (name1 name)
-	       (i 0)
-	       ;; We do not want to run timers.
-	       timer-list timer-idle-list)
+	       (i 0))
 
 	  (while (get-process name1)
 	    ;; NAME must be unique as process name.
@@ -1092,23 +995,16 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
 		    ;; otherwise we might be interrupted by
 		    ;; `verify-visited-file-modtime'.
 		    (let ((buffer-undo-list t)
-			  (buffer-read-only nil)
-			  (mark (point)))
+			  (inhibit-read-only t))
 		      (clear-visited-file-modtime)
 		      (narrow-to-region (point-max) (point-max))
 		      ;; We call `tramp-adb-maybe-open-connection', in
 		      ;; order to cleanup the prompt afterwards.
 		      (tramp-adb-maybe-open-connection v)
-		      (widen)
-		      (delete-region mark (point))
-		      (narrow-to-region (point-max) (point-max))
+		      (delete-region (point-min) (point-max))
 		      ;; Send the command.
-		      (let ((tramp-adb-prompt (regexp-quote command)))
-			(tramp-adb-send-command v command))
-		      (let ((p (tramp-get-connection-process v)))
-			;; Stop process if indicated.
-			(when stop
-			  (stop-process p))
+		      (let* ((p (tramp-get-connection-process v)))
+                        (tramp-adb-send-command v command nil t) ; nooutput
 			;; Set sentinel and filter.
 			(when sentinel
 			  (set-process-sentinel p sentinel))
@@ -1120,6 +1016,14 @@ PRESERVE-UID-GID and PRESERVE-EXTENDED-ATTRIBUTES are completely ignored."
 			(ignore-errors
 			  (set-process-query-on-exit-flag p (null noquery))
 			  (set-marker (process-mark p) (point)))
+			;; Read initial output.  Remove the first line,
+			;; which is the command echo.
+			(while
+			    (progn
+			      (goto-char (point-min))
+			      (not (re-search-forward "[\n]" nil t)))
+			  (tramp-accept-process-output p 0))
+			(delete-region (point-min) (point))
 			;; Return process.
 			p))))
 
@@ -1158,7 +1062,7 @@ E.g. a host name \"192.168.1.1#5555\" returns \"192.168.1.1:5555\"
   (with-tramp-connection-property (tramp-get-connection-process vec) "device"
     (let* ((host (tramp-file-name-host vec))
 	   (port (tramp-file-name-port-or-default vec))
-	   (devices (mapcar 'cadr (tramp-adb-parse-device-names nil))))
+	   (devices (mapcar #'cadr (tramp-adb-parse-device-names nil))))
       (replace-regexp-in-string
        tramp-prefix-port-format ":"
        (cond ((member host devices) host)
@@ -1195,7 +1099,7 @@ E.g. a host name \"192.168.1.1#5555\" returns \"192.168.1.1:5555\"
     (prog1
 	(unless
 	    (zerop
-	     (apply 'tramp-call-process vec tramp-adb-program nil t nil args))
+	     (apply #'tramp-call-process vec tramp-adb-program nil t nil args))
 	  (buffer-string))
       (tramp-message vec 6 "%s" (buffer-string)))))
 
@@ -1207,24 +1111,27 @@ This happens for Android >= 4.0."
 
 ;; Connection functions
 
-(defun tramp-adb-send-command (vec command)
+(defun tramp-adb-send-command (vec command &optional neveropen nooutput)
   "Send the COMMAND to connection VEC."
-  (tramp-adb-maybe-open-connection vec)
+  (unless neveropen (tramp-adb-maybe-open-connection vec))
   (tramp-message vec 6 "%s" command)
   (tramp-send-string vec command)
-  ;; fixme: Race condition
-  (tramp-adb-wait-for-output (tramp-get-connection-process vec))
-  (with-current-buffer (tramp-get-connection-buffer vec)
-    (save-excursion
-      (goto-char (point-min))
-      ;; We can't use stty to disable echo of command.
-      (delete-matching-lines (regexp-quote command))
-      ;; When the local machine is W32, there are still trailing ^M.
-      ;; There must be a better solution by setting the correct coding
-      ;; system, but this requires changes in core Tramp.
-      (goto-char (point-min))
-      (while (re-search-forward "\r+$" nil t)
-	(replace-match "" nil nil)))))
+  (unless nooutput
+    ;; FIXME: Race condition.
+    (tramp-adb-wait-for-output (tramp-get-connection-process vec))
+    (with-current-buffer (tramp-get-connection-buffer vec)
+      (save-excursion
+	(goto-char (point-min))
+	;; We can't use stty to disable echo of command.  stty is said
+	;; to be added to toybox 0.7.6.  busybox shall have it, but this
+	;; isn't used any longer for Android.
+	(delete-matching-lines (regexp-quote command))
+	;; When the local machine is W32, there are still trailing ^M.
+	;; There must be a better solution by setting the correct coding
+	;; system, but this requires changes in core Tramp.
+	(goto-char (point-min))
+	(while (re-search-forward "\r+$" nil t)
+	  (replace-match "" nil nil))))))
 
 (defun tramp-adb-send-command-and-check (vec command)
   "Run COMMAND and check its exit status.
@@ -1243,51 +1150,43 @@ the exit status is not equal 0, and t otherwise."
     (skip-chars-forward "^ ")
     (prog1
 	(zerop (read (current-buffer)))
-      (let (buffer-read-only)
+      (let ((inhibit-read-only t))
 	(delete-region (match-beginning 0) (point-max))))))
 
 (defun tramp-adb-barf-unless-okay (vec command fmt &rest args)
   "Run COMMAND, check exit status, throw error if exit status not okay.
 FMT and ARGS are passed to `error'."
   (unless (tramp-adb-send-command-and-check vec command)
-    (apply 'tramp-error vec 'file-error fmt args)))
+    (apply #'tramp-error vec 'file-error fmt args)))
 
 (defun tramp-adb-wait-for-output (proc &optional timeout)
   "Wait for output from remote command."
   (unless (buffer-live-p (process-buffer proc))
     (delete-process proc)
     (tramp-error proc 'file-error "Process `%s' not available, try again" proc))
-  (with-current-buffer (process-buffer proc)
-    (if (tramp-wait-for-regexp
-	 proc timeout
-	 (tramp-get-connection-property proc "prompt" tramp-adb-prompt))
-	(let (buffer-read-only)
-	  (goto-char (point-min))
-	  ;; ADB terminal sends "^H" sequences.
-	  (when (re-search-forward "<\b+" (point-at-eol) t)
-	    (forward-line 1)
-	    (delete-region (point-min) (point)))
-	  ;; Delete the prompt.
-         (goto-char (point-min))
-         (when (re-search-forward
-		(tramp-get-connection-property proc "prompt" tramp-adb-prompt)
-		(point-at-eol) t)
-           (forward-line 1)
-           (delete-region (point-min) (point)))
-	  (goto-char (point-max))
-	  (re-search-backward
-	   (tramp-get-connection-property proc "prompt" tramp-adb-prompt) nil t)
-	  (delete-region (point) (point-max)))
-      (if timeout
+  (let ((prompt (tramp-get-connection-property proc "prompt" tramp-adb-prompt)))
+    (with-current-buffer (process-buffer proc)
+      (if (tramp-wait-for-regexp proc timeout prompt)
+	  (let ((inhibit-read-only t))
+	    (goto-char (point-min))
+	    ;; ADB terminal sends "^H" sequences.
+	    (when (re-search-forward "<\b+" (point-at-eol) t)
+	      (forward-line 1)
+	      (delete-region (point-min) (point)))
+	    ;; Delete the prompt.
+            (goto-char (point-min))
+            (when (re-search-forward prompt (point-at-eol) t)
+              (forward-line 1)
+              (delete-region (point-min) (point)))
+	    (goto-char (point-max))
+	    (re-search-backward prompt nil t)
+	    (delete-region (point) (point-max)))
+	(if timeout
+	    (tramp-error
+	     proc 'file-error
+	     "[[Remote prompt `%s' not found in %d secs]]" prompt timeout)
 	  (tramp-error
-	   proc 'file-error
-	   "[[Remote adb prompt `%s' not found in %d secs]]"
-	   (tramp-get-connection-property proc "prompt" tramp-adb-prompt)
-	   timeout)
-	(tramp-error
-	 proc 'file-error
-	 "[[Remote prompt `%s' not found]]"
-	 (tramp-get-connection-property proc "prompt" tramp-adb-prompt))))))
+	   proc 'file-error "[[Remote prompt `%s' not found]]" prompt))))))
 
 (defun tramp-adb-maybe-open-connection (vec)
   "Maybe open a connection VEC.
@@ -1306,6 +1205,14 @@ connection if a previous connection has died for some reason."
       (tramp-error vec 'file-error "Cannot switch to user `%s'" user))
 
     (unless (process-live-p p)
+      ;; During completion, don't reopen a new connection.  We check
+      ;; this for the process related to `tramp-buffer-name';
+      ;; otherwise `start-file-process' wouldn't run ever when
+      ;; `non-essential' is non-nil.
+      (when (and (tramp-completion-mode-p)
+		 (null (get-process (tramp-buffer-name vec))))
+	(throw 'non-essential 'non-essential))
+
       (save-match-data
 	(when (and p (processp p)) (delete-process p))
 	(if (zerop (length device))
@@ -1318,18 +1225,24 @@ connection if a previous connection has died for some reason."
 			 (list "shell")))
 		 (p (let ((default-directory
 			    (tramp-compat-temporary-file-directory)))
-		      (apply 'start-process (tramp-get-connection-name vec) buf
+		      (apply #'start-process (tramp-get-connection-name vec) buf
 			     tramp-adb-program args)))
 		 (prompt (md5 (concat (prin1-to-string process-environment)
 				      (current-time-string)))))
 	    (tramp-message
-	     vec 6 "%s" (mapconcat 'identity (process-command p) " "))
-	    ;; Wait for initial prompt.
+	     vec 6 "%s" (string-join (process-command p) " "))
+	    ;; Wait for initial prompt.  On some devices, it needs an
+	    ;; initial RET, in order to get it.
+            (sleep-for 0.1)
+	    (tramp-send-string vec tramp-rsh-end-of-line)
 	    (tramp-adb-wait-for-output p 30)
 	    (unless (process-live-p p)
-	      (tramp-error  vec 'file-error "Terminated!"))
+	      (tramp-error vec 'file-error "Terminated!"))
+
+	    ;; Set sentinel and query flag.  Initialize variables.
+	    (set-process-sentinel p #'tramp-process-sentinel)
 	    (process-put p 'vector vec)
-	    (process-put p 'adjust-window-size-function 'ignore)
+	    (process-put p 'adjust-window-size-function #'ignore)
 	    (set-process-query-on-exit-flag p nil)
 
 	    ;; Change prompt.
@@ -1377,6 +1290,24 @@ connection if a previous connection has died for some reason."
 	    ;; Mark it as connected.
 	    (tramp-set-connection-property p "connected" t)))))))
 
+;; Default settings for connection-local variables.
+(defconst tramp-adb-connection-local-default-profile
+  '((shell-file-name . "/system/bin/sh")
+    (shell-command-switch . "-c"))
+  "Default connection-local variables for remote adb connections.")
+
+;; `connection-local-set-profile-variables' and
+;; `connection-local-set-profiles' exists since Emacs 26.1.
+(with-eval-after-load 'shell
+  (tramp-compat-funcall
+   'connection-local-set-profile-variables
+   'tramp-adb-connection-local-default-profile
+   tramp-adb-connection-local-default-profile)
+  (tramp-compat-funcall
+   'connection-local-set-profiles
+   `(:application tramp :protocol ,tramp-adb-method)
+   'tramp-adb-connection-local-default-profile))
+
 (add-hook 'tramp-unload-hook
 	  (lambda ()
 	    (unload-feature 'tramp-adb 'force)))
