@@ -1,6 +1,6 @@
 ;;; tramp-archive.el --- Tramp archive manager  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2017-2018 Free Software Foundation, Inc.
+;; Copyright (C) 2017-2019 Free Software Foundation, Inc.
 
 ;; Author: Michael Albinus <michael.albinus@gmx.de>
 ;; Keywords: comm, processes
@@ -77,8 +77,8 @@
 ;; File archives could also be compressed, identified by an additional
 ;; compression suffix.  Valid compression suffixes are listed in the
 ;; constant `tramp-archive-compression-suffixes'.  They are ".bz2",
-;; ".gz", ".lrz", ".lz", ".lz4", ".lzma", ".lzo", ".uu", ".xz" and
-;; ".Z".  A valid archive file name would be
+;; ".gz", ".lrz", ".lz", ".lz4", ".lzma", ".lzo", ".uu", ".xz", ".Z",
+;; and ".zst".  A valid archive file name would be
 ;; "/path/to/dir/file.tar.gz/dir/file".  Even several suffixes in a
 ;; row are possible, like "/path/to/dir/file.tar.gz.uu/dir/file".
 
@@ -108,7 +108,10 @@
 ;;; Code:
 
 (eval-when-compile (require 'cl-lib))
-(require 'tramp-gvfs)
+;; Sometimes, compilation fails with "Variable binding depth exceeds
+;; max-specpdl-size".
+(eval-and-compile
+  (let ((max-specpdl-size (* 2 max-specpdl-size))) (require 'tramp-gvfs)))
 
 (autoload 'dired-uncache "dired")
 (autoload 'url-tramp-convert-url-to-tramp "url-tramp")
@@ -126,6 +129,8 @@
 (setq tramp-archive-enabled tramp-gvfs-enabled)
 
 ;; <https://github.com/libarchive/libarchive/wiki/LibarchiveFormats>
+;; Note: "arc" and "zoo" are supported by `archive-mode', but they
+;; don't work here.
 ;;;###autoload
 (defconst tramp-archive-suffixes
   ;; "cab", "lzh", "msu" and "zip" are included with lower and upper
@@ -149,7 +154,7 @@
     "rar" ;; RAR archives.
     "rpm" ;; Red Hat packages.
     "shar" ;; Shell archives.  Not in libarchive testsuite.
-    "tar" "tbz" "tgz" "tlz" "txz" ;; (Compressed) tape archives.
+    "tar" "tbz" "tgz" "tlz" "txz" "tzst" ;; (Compressed) tape archives.
     "warc" ;; Web archives.
     "xar" ;; macOS XAR archives.  Not in libarchive testsuite.
     "xpi" ;; XPInstall Mozilla addons.  Not in libarchive testsuite.
@@ -164,7 +169,7 @@ It must be supported by libarchive(3).")
 
 ;;;###autoload
 (defconst tramp-archive-compression-suffixes
-  '("bz2" "gz" "lrz" "lz" "lz4" "lzma" "lzo" "uu" "xz" "Z")
+  '("bz2" "gz" "lrz" "lz" "lz4" "lzma" "lzo" "uu" "xz" "Z" "zst")
   "List of suffixes which indicate a compressed file.
 It must be supported by libarchive(3).")
 
@@ -198,15 +203,15 @@ It must be supported by libarchive(3).")
 (defconst tramp-archive-all-gvfs-methods
   (cons tramp-archive-method
 	(let ((values (cdr (cadr (get 'tramp-gvfs-methods 'custom-type)))))
-	  (setq values (mapcar 'last values)
-		values (mapcar 'car values))))
+	  (setq values (mapcar #'last values)
+		values (mapcar #'car values))))
   "List of all methods `tramp-gvfs-methods' offers.")
 
 
 ;; New handlers should be added here.
 ;;;###tramp-autoload
 (defconst tramp-archive-file-name-handler-alist
-  '((access-file . ignore)
+  '((access-file . tramp-archive-handle-access-file)
     (add-name-to-file . tramp-archive-handle-not-implemented)
     ;; `byte-compiler-base-file-name' performed by default handler.
     ;; `copy-directory' performed by default handler.
@@ -284,8 +289,9 @@ Operations not mentioned here will be handled by the default Emacs primitives.")
 
 (defsubst tramp-archive-file-name-for-operation (operation &rest args)
   "Like `tramp-file-name-for-operation', but for archive file name syntax."
-  (cl-letf (((symbol-function 'tramp-tramp-file-p) 'tramp-archive-file-name-p))
-    (apply 'tramp-file-name-for-operation operation args)))
+  (cl-letf (((symbol-function #'tramp-tramp-file-p)
+	     #'tramp-archive-file-name-p))
+    (apply #'tramp-file-name-for-operation operation args)))
 
 (defun tramp-archive-run-real-handler (operation args)
   "Invoke normal file name handler for OPERATION.
@@ -310,14 +316,15 @@ pass to the OPERATION."
           (tramp-register-file-name-handlers)
           (tramp-archive-run-real-handler operation args))
 
-      (let* ((filename (apply 'tramp-archive-file-name-for-operation
+      (let* ((filename (apply #'tramp-archive-file-name-for-operation
 			      operation args))
 	     (archive (tramp-archive-file-name-archive filename)))
 
-        ;; The file archive could be a directory, see Bug#30293.
-        (if (and archive
-	         (tramp-archive-run-real-handler
-                  'file-directory-p (list archive)))
+        ;; `filename' could be a quoted file name.  Or the file
+        ;; archive could be a directory, see Bug#30293.
+        (if (or (null archive)
+	        (tramp-archive-run-real-handler
+                 #'file-directory-p (list archive)))
             (tramp-archive-run-real-handler operation args)
           ;; Now run the handler.
           (let ((tramp-methods (cons `(,tramp-archive-method) tramp-methods))
@@ -326,7 +333,7 @@ pass to the OPERATION."
 	        (tramp-unknown-id-integer (user-uid))
 	        (tramp-unknown-id-string (user-login-name))
 	        (fn (assoc operation tramp-archive-file-name-handler-alist)))
-	    (when (eq (cdr fn) 'tramp-archive-handle-not-implemented)
+	    (when (eq (cdr fn) #'tramp-archive-handle-not-implemented)
 	      (setq args (cons operation args)))
 	    (if fn
 	        (save-match-data (apply (cdr fn) args))
@@ -334,7 +341,7 @@ pass to the OPERATION."
 
 ;;;###autoload
 (defalias
-  'tramp-archive-autoload-file-name-handler 'tramp-autoload-file-name-handler)
+  'tramp-archive-autoload-file-name-handler #'tramp-autoload-file-name-handler)
 
 ;;;###autoload
 (progn (defun tramp-register-archive-file-name-handler ()
@@ -342,17 +349,17 @@ pass to the OPERATION."
   (when tramp-archive-enabled
     (add-to-list 'file-name-handler-alist
 	         (cons (tramp-archive-autoload-file-name-regexp)
-		       'tramp-archive-autoload-file-name-handler))
+		       #'tramp-archive-autoload-file-name-handler))
     (put 'tramp-archive-autoload-file-name-handler 'safe-magic t))))
 
 ;;;###autoload
 (progn
-  (add-hook 'after-init-hook 'tramp-register-archive-file-name-handler)
+  (add-hook 'after-init-hook #'tramp-register-archive-file-name-handler)
   (add-hook
    'tramp-archive-unload-hook
    (lambda ()
      (remove-hook
-      'after-init-hook 'tramp-register-archive-file-name-handler))))
+      'after-init-hook #'tramp-register-archive-file-name-handler))))
 
 ;; In older Emacsen (prior 27.1), the autoload above does not exist.
 ;; So we call it again; it doesn't hurt.
@@ -360,19 +367,18 @@ pass to the OPERATION."
 
 ;; Mark `operations' the handler is responsible for.
 (put 'tramp-archive-file-name-handler 'operations
-     (mapcar 'car tramp-archive-file-name-handler-alist))
+     (mapcar #'car tramp-archive-file-name-handler-alist))
 
 ;; `tramp-archive-file-name-handler' must be placed before `url-file-handler'.
 (when url-handler-mode (tramp-register-file-name-handlers))
 
-(eval-after-load 'url-handler
-  (progn
-    (add-hook 'url-handler-mode-hook 'tramp-register-file-name-handlers)
-    (add-hook
-     'tramp-archive-unload-hook
-     (lambda ()
-       (remove-hook
-	'url-handler-mode-hook 'tramp-register-file-name-handlers)))))
+(with-eval-after-load 'url-handler
+  (add-hook 'url-handler-mode-hook #'tramp-register-file-name-handlers)
+  (add-hook
+   'tramp-archive-unload-hook
+   (lambda ()
+     (remove-hook
+      'url-handler-mode-hook #'tramp-register-file-name-handlers))))
 
 
 ;; File name conversions.
@@ -380,6 +386,8 @@ pass to the OPERATION."
 (defun tramp-archive-file-name-p (name)
   "Return t if NAME is a string with archive file name syntax."
   (and (stringp name)
+       ;; `tramp-archive-file-name-regexp' does not suppress quoted file names.
+       (not (tramp-compat-file-name-quoted-p name t))
        ;; We cannot use `string-match-p', the matches are used.
        (string-match tramp-archive-file-name-regexp name)
        t))
@@ -452,9 +460,9 @@ name is kept in slot `hop'"
 
        ;; Anything else.  Here we call `file-local-copy', which we
        ;; have avoided so far.
-       (t (let* ((inhibit-file-name-operation 'file-local-copy)
+       (t (let* ((inhibit-file-name-operation #'file-local-copy)
 		 (inhibit-file-name-handlers
-		  (cons 'jka-compr-handler inhibit-file-name-handlers))
+		  (cons #'jka-compr-handler inhibit-file-name-handlers))
 		 (copy (file-local-copy archive)))
 	    (setf (tramp-file-name-host vec) (tramp-archive-gvfs-host copy))
 	    (puthash archive (cons vec copy) tramp-archive-hash))))
@@ -464,26 +472,30 @@ name is kept in slot `hop'"
       (setf (tramp-file-name-localname vec) localname)
       vec)))
 
-;;;###tramp-autoload
 (defun tramp-archive-cleanup-hash ()
   "Remove local copies of archives, used by GVFS."
-  (maphash
-   (lambda (key value)
-     ;; Unmount local copy.
-     (ignore-errors
-       (tramp-message (car value) 3 "Unmounting %s" (or (cdr value) key))
-       (tramp-gvfs-unmount (car value)))
-     ;; Delete local copy.
-     (ignore-errors (delete-file (cdr value)))
-     (remhash key tramp-archive-hash))
-   tramp-archive-hash)
-  (clrhash tramp-archive-hash))
+  ;; Don't check for a proper method.
+  (let ((non-essential t))
+    (maphash
+     (lambda (key value)
+       ;; Unmount local copy.
+       (ignore-errors
+	 (tramp-message (car value) 3 "Unmounting %s" (or (cdr value) key))
+	 (tramp-gvfs-unmount (car value)))
+       ;; Delete local copy.
+       (ignore-errors (delete-file (cdr value)))
+       (remhash key tramp-archive-hash))
+     tramp-archive-hash)
+    (clrhash tramp-archive-hash)))
 
-(add-hook 'kill-emacs-hook 'tramp-archive-cleanup-hash)
+(add-hook 'tramp-cleanup-all-connections-hook #'tramp-archive-cleanup-hash)
+(add-hook 'kill-emacs-hook #'tramp-archive-cleanup-hash)
 (add-hook 'tramp-archive-unload-hook
 	  (lambda ()
+	    (remove-hook 'tramp-cleanup-all-connections-hook
+			 #'tramp-archive-cleanup-hash)
 	    (remove-hook 'kill-emacs-hook
-			 'tramp-archive-cleanup-hash)))
+			 #'tramp-archive-cleanup-hash)))
 
 (defsubst tramp-file-name-archive (vec)
   "Extract the archive file name from VEC.
@@ -528,6 +540,10 @@ offered."
 
 ;; File name primitives.
 
+(defun tramp-archive-handle-access-file (filename string)
+  "Like `access-file' for Tramp files."
+  (access-file (tramp-archive-gvfs-file-name filename) string))
+
 (defun tramp-archive-handle-copy-file
   (filename newname &optional ok-if-already-exists keep-date
    preserve-uid-gid preserve-extended-attributes)
@@ -574,9 +590,7 @@ offered."
 
 (defun tramp-archive-handle-file-readable-p (filename)
   "Like `file-readable-p' for file archives."
-  (with-parsed-tramp-file-name
-      (tramp-archive-gvfs-file-name filename) nil
-    (tramp-check-cached-permissions v ?r)))
+  (file-readable-p (tramp-archive-gvfs-file-name filename)))
 
 (defun tramp-archive-handle-file-system-info (filename)
   "Like `file-system-info' for file archives."
@@ -629,7 +643,7 @@ offered."
   "Generic handler for operations not implemented for file archives."
   (let ((v (ignore-errors
 	     (tramp-archive-dissect-file-name
-	      (apply 'tramp-archive-file-name-for-operation operation args)))))
+	      (apply #'tramp-archive-file-name-for-operation operation args)))))
     (tramp-message v 10 "%s" (cons operation args))
     (tramp-error
      v 'file-error
