@@ -1,6 +1,6 @@
 ;;; tramp-ftp.el --- Tramp convenience functions for Ange-FTP  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2002-2018 Free Software Foundation, Inc.
+;; Copyright (C) 2002-2019 Free Software Foundation, Inc.
 
 ;; Author: Michael Albinus <michael.albinus@gmx.de>
 ;; Keywords: comm, processes
@@ -54,9 +54,8 @@ present for backward compatibility."
     (setq file-name-handler-alist
 	  (delete a1 (delete a2 file-name-handler-alist)))))
 
-(eval-after-load "ange-ftp"
-  '(when (functionp 'tramp-disable-ange-ftp)
-     (tramp-disable-ange-ftp)))
+(with-eval-after-load 'ange-ftp
+  (tramp-disable-ange-ftp))
 
 ;;;###tramp-autoload
 (defun tramp-ftp-enable-ange-ftp ()
@@ -87,7 +86,7 @@ present for backward compatibility."
 			 ange-ftp-completion-hook-function)
 		       file-name-handler-alist)))))
 
-(add-hook 'tramp-ftp-unload-hook 'tramp-ftp-enable-ange-ftp)
+(add-hook 'tramp-ftp-unload-hook #'tramp-ftp-enable-ange-ftp)
 
 ;; Define FTP method ...
 ;;;###tramp-autoload
@@ -131,7 +130,7 @@ pass to the OPERATION."
 	  (ange-ftp-ftp-name-arg "")
 	  (ange-ftp-ftp-name-res nil)
 	  (v (tramp-dissect-file-name
-	      (apply 'tramp-file-name-for-operation operation args) t)))
+	      (apply #'tramp-file-name-for-operation operation args) t)))
       (setf (tramp-file-name-method v) tramp-ftp-method)
       ;; Set "process-name" for thread support.
       (tramp-set-connection-property
@@ -149,7 +148,7 @@ pass to the OPERATION."
        ;; because this returns another user but the one declared in
        ;; "~/.netrc".
        ((memq operation '(file-directory-p file-exists-p))
-	(if (apply 'ange-ftp-hook-function operation args)
+	(if (apply #'ange-ftp-hook-function operation args)
 	    (tramp-set-connection-property v "started" t)
 	  nil))
 
@@ -181,7 +180,7 @@ pass to the OPERATION."
 			(and (eq inhibit-file-name-operation operation)
 			     inhibit-file-name-handlers)))
 		 (inhibit-file-name-operation operation))
-	    (apply 'ange-ftp-hook-function operation args)))))))
+	    (apply #'ange-ftp-hook-function operation args)))))))
 
 ;; It must be a `defsubst' in order to push the whole code into
 ;; tramp-loaddefs.el.  Otherwise, there would be recursive autoloading.
