@@ -1,6 +1,6 @@
 ;;; tramp-rclone.el --- Tramp access functions to cloud storages  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2018 Free Software Foundation, Inc.
+;; Copyright (C) 2018-2019 Free Software Foundation, Inc.
 
 ;; Author: Michael Albinus <michael.albinus@gmx.de>
 ;; Keywords: comm, processes
@@ -50,16 +50,15 @@
 
 ;;;###tramp-autoload
 (tramp--with-startup
- (add-to-list
-  'tramp-methods
-  `(,tramp-rclone-method
-    (tramp-mount-args nil)
-    (tramp-copyto-args nil)
-    (tramp-moveto-args nil)
-    (tramp-about-args ("--full")))))
+ (add-to-list 'tramp-methods
+	      `(,tramp-rclone-method
+		(tramp-mount-args nil)
+		(tramp-copyto-args nil)
+		(tramp-moveto-args nil)
+		(tramp-about-args ("--full"))))
+
+ (add-to-list 'tramp-default-host-alist `(,tramp-rclone-method nil ""))
 
-;;;###tramp-autoload
-(tramp--with-startup
  (tramp-set-completion-function
   tramp-rclone-method '((tramp-rclone-parse-device-names ""))))
 
@@ -67,7 +66,7 @@
 ;; New handlers should be added here.
 ;;;###tramp-autoload
 (defconst tramp-rclone-file-name-handler-alist
-  '((access-file . ignore)
+  '((access-file . tramp-handle-access-file)
     (add-name-to-file . tramp-handle-add-name-to-file)
     ;; `byte-compiler-base-file-name' performed by default handler.
     ;; `copy-directory' performed by default handler.
@@ -171,26 +170,12 @@ pass to the OPERATION."
 (defun tramp-rclone-parse-device-names (_ignore)
   "Return a list of (nil host) tuples allowed to access."
   (with-tramp-connection-property nil "rclone-device-names"
-    (with-timeout (10)
-      (with-temp-buffer
-	;; `call-process' does not react on timer under MS Windows.
-	;; That's why we use `start-process'.
-	(let ((p (start-process
-		  tramp-rclone-program (current-buffer)
-		  tramp-rclone-program "listremotes"))
-	      (v (make-tramp-file-name :method tramp-rclone-method))
-	      result)
-	  (tramp-message v 6 "%s" (mapconcat 'identity (process-command p) " "))
-	  (process-put p 'adjust-window-size-function 'ignore)
-	  (set-process-query-on-exit-flag p nil)
-	  (while (process-live-p p)
-	    (accept-process-output p 0.1))
-	  (accept-process-output p 0.1)
-	  (tramp-message v 6 "\n%s" (buffer-string))
-	  (goto-char (point-min))
-	  (while (search-forward-regexp "^\\(\\S-+\\):$" nil t)
-	    (push (list nil (match-string 1)) result))
-	  result)))))
+    (delq nil
+	  (mapcar
+	   (lambda (line)
+	     (when (string-match "^\\(\\S-+\\):$" line)
+	       `(nil ,(match-string 1 line))))
+	   (tramp-process-lines nil tramp-rclone-program "listremotes")))))
 
 
 ;; File name primitives.
@@ -295,7 +280,7 @@ file names."
        'copy filename newname ok-if-already-exists keep-date
        preserve-uid-gid preserve-extended-attributes)
     (tramp-run-real-handler
-     'copy-file
+     #'copy-file
      (list filename newname ok-if-already-exists keep-date
 	   preserve-uid-gid preserve-extended-attributes))))
 
@@ -329,7 +314,7 @@ file names."
 	(when full
 	  (let ((local (concat "^" (regexp-quote (tramp-rclone-mount-point v))))
 		(remote (funcall (if (tramp-compat-file-name-quoted-p directory)
-				     'tramp-compat-file-name-quote 'identity)
+				     #'tramp-compat-file-name-quote #'identity)
 				 (file-remote-p directory))))
 	    (setq result
 		  (mapcar
@@ -343,7 +328,7 @@ file names."
 			      result)))
 	    (setq result (cons item result))))
 	;; Return result.
-	(if nosort result (sort result 'string<))))))
+	(if nosort result (sort result #'string<))))))
 
 (defun tramp-rclone-handle-file-attributes (filename &optional id-format)
   "Like `file-attributes' for Tramp files."
@@ -451,7 +436,7 @@ file names."
        'rename filename newname ok-if-already-exists
        'keep-date 'preserve-uid-gid)
     (tramp-run-real-handler
-     'rename-file (list filename newname ok-if-already-exists))))
+     #'rename-file (list filename newname ok-if-already-exists))))
 
 
 ;; File name conversions.
@@ -461,7 +446,7 @@ file names."
   (expand-file-name
    (concat
     tramp-temp-name-prefix (tramp-file-name-method vec)
-    "."  (tramp-file-name-host vec))
+    "." (tramp-file-name-host vec))
    (tramp-compat-temporary-file-directory)))
 
 (defun tramp-rclone-mounted-p (vec)
@@ -471,12 +456,12 @@ file names."
     ;; to cache a nil result.
     (or (tramp-get-connection-property
 	 (tramp-get-connection-process vec) "mounted" nil)
-	(tramp-set-connection-property
-	 (tramp-get-connection-process vec) "mounted"
-	 (let* ((default-directory temporary-file-directory)
-		(mount (shell-command-to-string "mount -t fuse.rclone")))
-	   (tramp-message vec 6 "%s" "mount -t fuse.rclone")
-	   (tramp-message vec 6 "\n%s" mount)
+	(let* ((default-directory temporary-file-directory)
+	       (mount (shell-command-to-string "mount -t fuse.rclone")))
+	  (tramp-message vec 6 "%s" "mount -t fuse.rclone")
+	  (tramp-message vec 6 "\n%s" mount)
+	  (tramp-set-connection-property
+	   (tramp-get-connection-process vec) "mounted"
 	   (when (string-match
 		  (format
 		   "^\\(%s:\\S-*\\)" (regexp-quote (tramp-file-name-host vec)))
@@ -515,7 +500,7 @@ file names."
       (let ((quoted (tramp-compat-file-name-quoted-p localname))
 	    (localname (tramp-compat-file-name-unquote localname)))
 	(funcall
-	 (if quoted 'tramp-compat-file-name-quote 'identity)
+	 (if quoted #'tramp-compat-file-name-quote #'identity)
 	 (expand-file-name
 	  (if (file-name-absolute-p localname)
 	      (substring localname 1) localname)
@@ -546,11 +531,19 @@ connection if a previous connection has died for some reason."
       (if (zerop (length host))
 	  (tramp-error vec 'file-error "Storage %s not connected" host))
 
+      ;; During completion, don't reopen a new connection.  We check
+      ;; this for the process related to `tramp-buffer-name';
+      ;; otherwise `start-file-process' wouldn't run ever when
+      ;; `non-essential' is non-nil.
+      (when (and (tramp-completion-mode-p)
+		 (null (get-process (tramp-buffer-name vec))))
+	(throw 'non-essential 'non-essential))
+
       ;; We need a process bound to the connection buffer.  Therefore,
       ;; we create a dummy process.  Maybe there is a better solution?
       (unless (get-buffer-process (tramp-get-connection-buffer vec))
 	(let ((p (make-network-process
-		  :name (tramp-buffer-name vec)
+		  :name (tramp-get-connection-name vec)
 		  :buffer (tramp-get-connection-buffer vec)
 		  :server t :host 'local :service t :noquery t)))
 	  (process-put p 'vector vec)
@@ -567,7 +560,7 @@ connection if a previous connection has died for some reason."
       ;; DESTINATION of `tramp-call-process'.
       (unless (tramp-rclone-mounted-p vec)
 	(apply
-	 'tramp-call-process
+	 #'tramp-call-process
 	 vec tramp-rclone-program nil 0 nil
 	 (delq nil
 	       `("mount" ,(concat host ":/")
@@ -575,7 +568,11 @@ connection if a previous connection has died for some reason."
 		 ;; This could be nil.
 		 ,(tramp-get-method-parameter vec 'tramp-mount-args))))
 	(while (not (file-exists-p (tramp-make-tramp-file-name vec 'localname)))
-	  (tramp-cleanup-connection vec 'keep-debug 'keep-password)))))
+	  (tramp-cleanup-connection vec 'keep-debug 'keep-password))
+
+	;; Mark it as connected.
+	(tramp-set-connection-property
+	 (tramp-get-connection-process vec) "connected" t))))
 
   ;; In `tramp-check-cached-permissions', the connection properties
   ;; "{uid,gid}-{integer,string}" are used.  We set them to proper values.
@@ -594,7 +591,7 @@ connection if a previous connection has died for some reason."
     (erase-buffer)
     (let ((flags (tramp-get-method-parameter
 		  vec (intern (format "tramp-%s-args" (car args))))))
-      (apply 'tramp-call-process
+      (apply #'tramp-call-process
 	     vec tramp-rclone-program nil t nil (append args flags)))))
 
 (add-hook 'tramp-unload-hook
