@@ -1,6 +1,6 @@
 ;;; tramp-sudoedit.el --- Functions for accessing under root permissions  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2018 Free Software Foundation, Inc.
+;; Copyright (C) 2018-2019 Free Software Foundation, Inc.
 
 ;; Author: Michael Albinus <michael.albinus@gmx.de>
 ;; Keywords: comm, processes
@@ -44,8 +44,8 @@
 (tramp--with-startup
  (add-to-list 'tramp-methods
               `(,tramp-sudoedit-method
-                (tramp-sudo-login      (("sudo") ("-u" "%u") ("-S") ("-H")
-			                ("-p" "Password:") ("--")))))
+                (tramp-sudo-login (("sudo") ("-u" "%u") ("-S") ("-H")
+			           ("-p" "Password:") ("--")))))
 
  (add-to-list 'tramp-default-user-alist '("\\`sudoedit\\'" nil "root"))
 
@@ -63,7 +63,7 @@ See `tramp-actions-before-shell' for more info.")
 
 ;;;###tramp-autoload
 (defconst tramp-sudoedit-file-name-handler-alist
-  '((access-file . ignore)
+  '((access-file . tramp-handle-access-file)
     (add-name-to-file . tramp-sudoedit-handle-add-name-to-file)
     (byte-compiler-base-file-name . ignore)
     ;; `copy-directory' performed by default handler.
@@ -232,7 +232,6 @@ absolute file names."
 	  (file-times (tramp-compat-file-attribute-modification-time
 		       (file-attributes filename)))
 	  (file-modes (tramp-default-file-modes filename))
-	  ;; `file-extended-attributes' exists since Emacs 24.4.
 	  (attributes (and preserve-extended-attributes
 			   (apply #'file-extended-attributes (list filename))))
 	  (sudoedit-operation
@@ -284,7 +283,6 @@ absolute file names."
 
 	;; Handle `preserve-extended-attributes'.  We ignore possible
 	;; errors, because ACL strings could be incompatible.
-	;; `set-file-extended-attributes' exists since Emacs 24.4.
 	(when attributes
 	  (ignore-errors
 	    (apply #'set-file-extended-attributes (list newname attributes))))
@@ -299,8 +297,7 @@ absolute file names."
 	  (with-parsed-tramp-file-name newname v2
 	    (tramp-flush-file-properties
 	     v2 (file-name-directory v2-localname))
-	    (tramp-flush-file-properties v2 v2-localname)
-	    (when (tramp-rclone-file-name-p newname))))))))
+	    (tramp-flush-file-properties v2 v2-localname)))))))
 
 (defun tramp-sudoedit-handle-copy-file
   (filename newname &optional ok-if-already-exists keep-date
@@ -315,7 +312,7 @@ absolute file names."
        'copy filename newname ok-if-already-exists keep-date
        preserve-uid-gid preserve-extended-attributes)
     (tramp-run-real-handler
-     'copy-file
+     #'copy-file
      (list filename newname ok-if-already-exists keep-date
 	   preserve-uid-gid preserve-extended-attributes))))
 
@@ -354,6 +351,8 @@ If the localname part of the given file name starts with \"/../\" then
 the result will be a local, non-Tramp, file name."
   ;; If DIR is not given, use `default-directory' or "/".
   (setq dir (or dir default-directory "/"))
+  ;; Handle empty NAME.
+  (when (zerop (length name)) (setq name "."))
   ;; Unless NAME is absolute, concat DIR and NAME.
   (unless (file-name-absolute-p name)
     (setq name (concat (file-name-as-directory dir) name)))
@@ -453,7 +452,7 @@ the result will be a local, non-Tramp, file name."
 	  (delq
 	   nil
 	   (mapcar
-	    (lambda (l) (and (not (string-match-p  "^[[:space:]]*$" l)) l))
+	    (lambda (l) (and (not (string-match-p "^[[:space:]]*$" l)) l))
 	    (split-string (buffer-string) "\n" 'omit)))))))))
 
 (defun tramp-sudoedit-handle-file-readable-p (filename)
@@ -543,7 +542,7 @@ the result will be a local, non-Tramp, file name."
    ;; Preserve trailing "/".
   (funcall
    (if (string-equal (file-name-nondirectory filename) "")
-       'file-name-as-directory 'identity)
+       #'file-name-as-directory #'identity)
    (with-parsed-tramp-file-name (expand-file-name filename) nil
      (tramp-make-tramp-file-name
       v
@@ -602,15 +601,17 @@ of the symlink.  If TARGET is a Tramp file, only the localname
 component is used as the target of the symlink."
   (if (not (tramp-tramp-file-p (expand-file-name linkname)))
       (tramp-run-real-handler
-       'make-symbolic-link (list target linkname ok-if-already-exists))
+       #'make-symbolic-link (list target linkname ok-if-already-exists))
 
     (with-parsed-tramp-file-name linkname nil
       ;; If TARGET is a Tramp name, use just the localname component.
-      (when (and (tramp-tramp-file-p target)
-		 (tramp-file-name-equal-p v (tramp-dissect-file-name target)))
-	(setq target
-	      (tramp-file-name-localname
-	       (tramp-dissect-file-name (expand-file-name target)))))
+      ;; Don't check for a proper method.
+      (let ((non-essential t))
+	(when (and (tramp-tramp-file-p target)
+		   (tramp-file-name-equal-p v (tramp-dissect-file-name target)))
+	  (setq target
+		(tramp-file-name-localname
+		 (tramp-dissect-file-name (expand-file-name target))))))
 
       ;; If TARGET is still remote, quote it.
       (if (tramp-tramp-file-p target)
@@ -657,8 +658,7 @@ component is used as the target of the symlink."
   (with-parsed-tramp-file-name (expand-file-name filename) nil
     (when (and (stringp acl-string) (tramp-sudoedit-remote-acl-p v))
       ;; Massage `acl-string'.
-      (setq acl-string
-	    (mapconcat #'identity (split-string acl-string "\n" 'omit) ","))
+      (setq acl-string (string-join (split-string acl-string "\n" 'omit) ","))
       (prog1
 	  (tramp-sudoedit-send-command
 	   v "setfacl" "-m"
@@ -675,7 +675,7 @@ component is used as the target of the symlink."
 	    (type (and (stringp (nth 2 context)) (nth 2 context)))
 	    (range (and (stringp (nth 3 context)) (nth 3 context))))
 	(when (tramp-sudoedit-send-command
-	       v  "chcon"
+	       v "chcon"
 	       (when user (format "--user=%s" user))
 	       (when role (format "--role=%s" role))
 	       (when type (format "--type=%s" type))
@@ -745,10 +745,11 @@ ID-FORMAT valid values are `string' and `integer'."
 
 ;; Used in `tramp-sudoedit-sudo-actions'.
 (defun tramp-sudoedit-action-sudo (proc vec)
-  "Check, whether a sudo process copy has finished."
+  "Check, whether a sudo process has finished.
+Remove unneeded output."
   ;; There might be pending output for the exit status.
-  (tramp-accept-process-output proc 0.1)
-  (when (not (process-live-p proc))
+  (unless (process-live-p proc)
+    (while (tramp-accept-process-output proc 0))
     ;; Delete narrowed region, it would be in the way reading a Lisp form.
     (goto-char (point-min))
     (widen)
@@ -768,15 +769,27 @@ connection if a previous connection has died for some reason."
   ;; We need a process bound to the connection buffer.  Therefore, we
   ;; create a dummy process.  Maybe there is a better solution?
   (unless (tramp-get-connection-process vec)
+
+    ;; During completion, don't reopen a new connection.  We check
+    ;; this for the process related to `tramp-buffer-name'; otherwise
+    ;; `start-file-process' wouldn't run ever when `non-essential' is
+    ;; non-nil.
+    (when (and (tramp-completion-mode-p)
+	       (null (get-process (tramp-buffer-name vec))))
+      (throw 'non-essential 'non-essential))
+
     (let ((p (make-network-process
-	      :name (tramp-buffer-name vec)
+	      :name (tramp-get-connection-name vec)
 	      :buffer (tramp-get-connection-buffer vec)
 	      :server t :host 'local :service t :noquery t)))
       (process-put p 'vector vec)
       (set-process-query-on-exit-flag p nil)
 
       ;; Set connection-local variables.
-      (tramp-set-connection-local-variables vec))
+      (tramp-set-connection-local-variables vec)
+
+      ;; Mark it as connected.
+      (tramp-set-connection-property p "connected" t))
 
     ;; In `tramp-check-cached-permissions', the connection properties
     ;; "{uid,gid}-{integer,string}" are used.  We set them to proper values.
@@ -814,7 +827,7 @@ in case of error, t otherwise."
 	   (tramp-verbose (if (= tramp-verbose 3) 2 tramp-verbose))
 	   ;; We do not want to save the password.
 	   auth-source-save-behavior)
-      (tramp-message vec 6 "%s" (mapconcat #'identity (process-command p) " "))
+      (tramp-message vec 6 "%s" (string-join (process-command p) " "))
       ;; Avoid process status message in output buffer.
       (set-process-sentinel p #'ignore)
       (process-put p 'vector vec)
